function result = Rqa(x, param)
% result = Rqa(x, param)
%
% parameter:
%
% x                     (n,2) matrix of x,y fixations or
%                       (n,1) column vector of categories
% param.linelength      threshold linelength, e.g. 2
% param.radius          e.g. 64. For categories use 0.1
% param.mincluster      e.g. 8, minimum size of clusters in recurrence matrix
%
% result:
%
% result.n              n (length of x)
% result.recmat         recurrence matrix
% result.nrec           number of recurrences
% result.rec            percentage of recurrences
% result.det            determinism
% result.revdet         reverse determinism (see below)
% result.meanline       meanline
% result.maxline        maxline
% result.ent            entropy
% result.relent         relative entropy
% result.lam            laminarity
% result.tt             TT
% result.corm           center of recurrence mass
% result.clusters       number of recurrence clusters / nTriangle
%
% Author                Walter F. Bischof
% Date                  Original: September 2011 - September 2012
%                       Simplified version: August 2015
%                       Recurrence clusters: February 2018

% Default return values

result.n = NaN;
result.recmat = NaN;
result.nrec = NaN;
result.rec = NaN;
result.det = NaN;
result.revdet = NaN;
result.meanline = NaN;
result.maxline = NaN;
result.ent = NaN;
result.relent = NaN;
result.lam = NaN;
result.tt = NaN;
result.corm = NaN;
result.clusters = NaN;

n = size(x, 1);
result.n = n;

if n <= 1
    return
end

% Distance matrix.
distanceMatrix = ComputeDistanceMatrix(x, x);
nTriangle = n * (n - 1) / 2;

% Recurrence matrix
recurrenceMatrix = distanceMatrix <= param.radius;
result.recmat = recurrenceMatrix;

% Diagonal not needed for most measures
partialRecurrenceMatrix = triu(recurrenceMatrix, 1);

% Diagonals of recurrence matrix, excluding main diagonal
[recurrenceDiagonals, diagonals] = spdiags(partialRecurrenceMatrix);

% Global measures: recurrence
nRec = sum(sum(partialRecurrenceMatrix));
result.nrec = nRec;
result.rec = 100.0 * nRec / nTriangle;

if nRec == 0
    return
end

% Diagonal line measures: determinism, meanLine, maxLine,
% entropy, relative entropy
nDiagonals = length(diagonals);
thresholdedDiagonals = [];
for i = 1:nDiagonals
    d = [0; recurrenceDiagonals(:, i); 0];
    dchange = diff(d);
    dlength = find(dchange == -1) - find(dchange == 1);
    dlength = dlength(dlength >= param.linelength);
    thresholdedDiagonals = [thresholdedDiagonals; dlength];
end

if ~isempty(thresholdedDiagonals)
    result.det = 100 * sum(thresholdedDiagonals) / nRec;
    result.meanline = mean(thresholdedDiagonals);
    result.maxline = max(thresholdedDiagonals);
    [result.ent, result.relent] = ...
        Entropy(thresholdedDiagonals, param.linelength, result.maxline);
    result.corm = ...
        100 * sum(sum(recurrenceDiagonals)' .* diagonals) / ((n - 1) * nRec);
end

% Reverse det is like det but on the minor diagonal.
% Revdet indicates a scanpath that is repeated in the opposite direction.

flipPartialRecurrenceMatrix = flipud(triu(recurrenceMatrix, 1));
[flipRecurrenceDiagonals, flipDiagonals] = spdiags(flipPartialRecurrenceMatrix);
nFlipDiagonals = length(flipDiagonals);
thresholdedFlipDiagonals = [];
for i = 1:nFlipDiagonals
    d = [0; flipRecurrenceDiagonals(:, i); 0];
    dchange = diff(d);
    dlength = find(dchange == -1) - find(dchange == 1);
    dlength = dlength(dlength >= param.linelength);
    thresholdedFlipDiagonals = [thresholdedFlipDiagonals; dlength];
end
if ~isempty(thresholdedFlipDiagonals)
    result.revdet = 100 * sum(thresholdedFlipDiagonals) / nRec;
end

% horizontal+vertical line measures: laminarity and TT.
% These are slightly different from Zbilut & Webber and Marwan
% #(verticals + horizontals) in upper triangle =
% #(verticals) in upper + #(verticals) in lower triangle
% Set diagonal to zero to avoid counting over the diagonal

thresholdedVerticals = [];
paddedRM = [zeros(n, 1), recurrenceMatrix - eye(n), zeros(n, 1)];
% In CRA, diagonal elements can be zero
paddedRM = max(paddedRM, 0);
for i = 1:n
    v = paddedRM(i, :);
    vchange = diff(v);
    vlength = find(vchange == -1) - find(vchange == 1);
    vlength = vlength(vlength >= param.linelength);
    thresholdedVerticals = [thresholdedVerticals, vlength];
end

if ~isempty(thresholdedVerticals)
    result.lam = 100 * sum(thresholdedVerticals) / sum(sum(recurrenceMatrix));
    result.tt = mean(thresholdedVerticals);
end

% Recurrence clusters
if ~isfield(param, 'mincluster')
    param.mincluster = 8;
end
result.clusters = ...
    RecurrenceClusters(recurrenceMatrix, param.mincluster, nTriangle);
end

function [entropy, relativeEntropy] = Entropy(a, minLine, maxLine)
% Compute entropy and relative entropy
[p, ~] = hist(a, unique(a));
p = p / sum(p);
entropy = -sum(p.*log2(p));
disp([ p entropy]);
if maxLine == minLine
    relativeEntropy = NaN;
else
    relativeEntropy = entropy / log2(maxLine-minLine+1);
end
end

function distanceMatrix = ComputeDistanceMatrix(x, y)
% Compute the distance matrix between x and y.
distanceMatrix = pdist2(x, y);
end

function clusters = RecurrenceClusters(recmat, threshold, nTriangle)
% Compute number of recurrence clusters normalized by nTriangle
cc = bwconncomp(triu(recmat, 1));
npixels = cellfun(@numel, cc.PixelIdxList);
clusters = sum(npixels(npixels >= threshold)) * 100 / nTriangle;
end
